import asyncio
import datetime
import inspect
import io
import logging
import re
from typing import (Any, AsyncGenerator, Dict, List, Literal, Optional, Tuple,
                    Union, cast)

from bson import ObjectId
from PIL import Image

from nawah import data as Data
from nawah.classes import (ANALYTIC, ATTR, EXTN, METHOD,
                           NAWAH_DOC, NAWAH_ENV, NAWAH_EVENTS, NAWAH_QUERY,
                           ON_HANDLER_RETURN, PRE_HANDLER_RETURN,
                           ConvertAttrException, InvalidAttrException,
                           InvalidAttrTypeArgException,
                           InvalidAttrTypeException, MethodException,
                           MissingAttrException, Query, RESULTS)
from nawah.enums import DELETE_STRATEGY, Event
from nawah.utils import (_expand_attr, _set_none_value, _update_attr_values,
                         validate_doc)

from ._diff_attrs import diff_attrs

logger = logging.getLogger('nawah')


class BaseModule:
    _nawah_module: bool = True

    collection: Optional[str]
    attrs: Dict[str, ATTR]
    diff: Union[bool, ATTR]
    create_draft: Union[bool, ATTR]
    update_draft: Union[bool, ATTR]
    defaults: Dict[str, Any]
    unique_attrs: List[Union[str, Tuple[str, ...]]]
    extns: Dict[str, Union[EXTN, ATTR]]
    privileges: List[str]
    methods: Dict[str, METHOD]
    analytics: List[ANALYTIC]

    package_name: str
    module_name: str

    def __init__(self):
        if not getattr(self, 'collection', None):
            self.collection = None
        if not getattr(self, 'attrs', None):
            self.attrs = {}
        if not getattr(self, 'diff', None):
            self.diff = False
        if not getattr(self, 'create_draft', None):
            self.create_draft = False
        if not getattr(self, 'update_draft', None):
            self.update_draft = False
        if not getattr(self, 'defaults', None):
            self.defaults = {}
        if not getattr(self, 'unique_attrs', None):
            self.unique_attrs = []
        if not getattr(self, 'extns', None):
            self.extns = {}
        if not getattr(self, 'privileges', None):
            self.privileges = ['read', 'create', 'update', 'delete', 'admin']
        if not getattr(self, 'methods', None):
            self.methods = {}
        if not getattr(self, 'cache', None):
            self.cache = []
        if not getattr(self, 'analytics', None):
            self.analytics = []

        # [DOC] Populate package and module names for in-context use.
        self.package_name = (
            self.__module__.replace('modules.', '').upper().split('.')[-2]
        )
        self.module_name = re.sub(
            r'([A-Z])',
            r'_\1',
            self.__class__.__name__[0].lower() + self.__class__.__name__[1:],
        ).lower()

    def _pre_initialise(self) -> None:
        pass

    def _initialise(self) -> None:
        # [DOC] Call _pre_initialise for advanced module initialisation
        self._pre_initialise()
        # [DOC] Check attrs for any invalid type
        for attr in self.attrs.keys():
            try:
                logger.debug(
                    f'Attempting to validate Attr Type for \'{attr}\' of module \'{self.module_name}\'.'
                )
                ATTR.validate_type(attr_type=self.attrs[attr])
            except InvalidAttrTypeException as e:
                logger.error(
                    f'Invalid Attr Type for \'{attr}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                )
                exit(1)
            except InvalidAttrTypeArgException as e:
                logger.error(
                    f'Invalid Attr Type Arg for \'{attr}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                )
                exit(1)
            # [DOC] Check default for invalid types, update default value
            for default in self.defaults.keys():
                if (
                    default == attr
                    or default.startswith(f'{attr}.')
                    or default.startswith(f'{attr}:')
                ):
                    if type(self.defaults[default]) == ATTR:
                        if self.defaults[default]._type != 'TYPE':
                            logger.error(
                                f'Invalid Attr Type for default \'{default}\' of module \'{self.module_name}\'. Only Attr Type TYPE is allowed. Exiting.'
                            )
                            exit(1)
                        logger.debug(
                            f'Attempting to validate Attr Type of default \'{default}\' of module \'{self.module_name}\'.'
                        )
                        try:
                            ATTR.validate_type(attr_type=self.defaults[default])
                        except InvalidAttrTypeException as e:
                            logger.error(
                                f'Invalid Attr Type for default \'{default}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                            )
                            exit(1)
                        except InvalidAttrTypeArgException as e:
                            logger.error(
                                f'Invalid Attr Type Arg for default \'{default}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                            )
                            exit(1)
                    logger.debug(
                        f'Updating default value for attr \'{attr}\' to: \'{self.defaults[default]}\''
                    )
                    _update_attr_values(
                        attr=ATTR.TYPED_DICT(dict=self.attrs),
                        value='default',
                        value_path=default,
                        value_val=self.defaults[default],
                    )
            # [DOC] Update extn value
            for extn in self.extns.keys():
                if (
                    extn == attr
                    or extn.startswith(f'{attr}.')
                    or extn.startswith(f'{attr}:')
                ):
                    logger.debug(
                        f'Updating extn value for attr \'{extn}\' to: \'{self.extns[extn]}\''
                    )
                    _update_attr_values(
                        attr=ATTR.TYPED_DICT(dict=self.attrs),
                        value='extn',
                        value_path=extn,
                        value_val=self.extns[extn],
                    )
        # [DOC] Abstract methods as BaseMethod objects
        for method_name in self.methods.keys():
            method = self.methods[method_name]
            # [DOC] Check value type
            if type(method) != METHOD:
                logger.error(
                    f'Invalid method \'{method}\' of module \'{self.module_name}\'. Exiting.'
                )
                exit(1)
            # [DOC] Set Method._module value to create back-link from child to parent
            method._module = self
            method._method_name = method_name
            # [DOC] Check valid method (and initialise it)
            try:
                method._validate()
            except Exception as e:
                logger.error(e)
                logger.error('Exiting.')
                exit(1)
        # [DOC] Check extns for invalid extended attrs
        for attr in self.extns.keys():
            if type(self.extns[attr]) not in [EXTN, ATTR]:
                logger.error(
                    f'Invalid extns attr \'{attr}\' of module \'{self.module_name}\'. Exiting.'
                )
                exit(1)
            if type(self.extns[attr]) == ATTR:
                self.extns[attr] = cast(ATTR, self.extns[attr])
                if self.extns[attr]._type != 'TYPE':
                    logger.error(
                        f'Invalid Attr Type for extn \'{attr}\' of module \'{self.module_name}\'. Only Attr Type TYPE is allowed. Exiting.'
                    )
                    exit(1)
                logger.debug(
                    f'Attempting to validate Attr Type of extn \'{attr}\' of module \'{self.module_name}\'.'
                )
                try:
                    ATTR.validate_type(attr_type=self.extns[attr])
                except InvalidAttrTypeException as e:
                    logger.error(
                        f'Invalid Attr Type for extn \'{attr}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                    )
                    exit(1)
                except InvalidAttrTypeArgException as e:
                    logger.error(
                        f'Invalid Attr Type Arg for extn \'{attr}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                    )
                    exit(1)

        # [DOC] Check valid type, value for diff
        if type(self.diff) not in [bool, ATTR]:
            logger.error(f'Invalid diff for module \'{self.module_name}\'. Exiting.')
            exit(1)
        if type(self.diff) == ATTR:
            self.diff = cast(ATTR, self.diff)
            if self.diff._type != 'TYPE':
                logger.error(
                    f'Invalid Attr Type for diff of module \'{self.module_name}\'. Only Attr Type TYPE is allowed. Exiting.'
                )
                exit(1)
            logger.debug(
                f'Attempting to validate Attr Type diff of module \'{self.module_name}\'.'
            )
            try:
                ATTR.validate_type(attr_type=self.diff)
            except InvalidAttrTypeException as e:
                logger.error(
                    f'Invalid Attr Type for diff of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                )
                exit(1)
            except InvalidAttrTypeArgException as e:
                logger.error(
                    f'Invalid Attr Type Arg for diff of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                )
                exit(1)

        # [DOC] Check valid types, values for create_draft, update_draft
        for attr in ['create_draft', 'update_draft']:
            if type(getattr(self, attr)) not in [bool, ATTR]:
                logger.error(
                    f'Invalid {attr} for module \'{self.module_name}\'. Exiting.'
                )
                exit(1)
            if type(getattr(self, attr)) == ATTR:
                if getattr(self, attr)._type != 'TYPE':
                    logger.error(
                        f'Invalid Attr Type for {attr} of module \'{self.module_name}\'. Only Attr Type TYPE is allowed. Exiting.'
                    )
                    exit(1)
                logger.debug(
                    f'Attempting to validate Attr Type {attr} of module \'{self.module_name}\'.'
                )
                try:
                    ATTR.validate_type(attr_type=getattr(self, attr))
                except InvalidAttrTypeException as e:
                    logger.error(
                        f'Invalid Attr Type for {attr} of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                    )
                    exit(1)
                except InvalidAttrTypeArgException as e:
                    logger.error(
                        f'Invalid Attr Type Arg for {attr} of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
                    )
                    exit(1)

        logger.debug(f'Initialised module {self.module_name}')

    def status(
        self,
        *,
        status: int,
        msg: str,
        args: Optional[Union[Dict[str, Any]]] = None,
    ) -> RESULTS:
        if status != 200:
            logger.warning(
                f'BaseModule.status with msg \'{msg}\' is not called with \'200\' status code. Use BaseModule.exception instead.'
            )

        status_dict = {'status': status, 'msg': msg, 'args': {}}
        if args:
            if 'code' in args.keys():
                args[
                    'code'
                ] = f'{self.package_name.upper()}_{self.module_name.upper()}_{args["code"]}'
            status_dict['args'] = args
        return status_dict

    def exception(
        self,
        *,
        status: int,
        msg: str,
        args: Optional[Union[Dict[str, Any]]] = None,
    ) -> MethodException:
        if status == 200:
            logger.warning(
                f'BaseModule.exception with msg \'{msg}\' is called with \'200\' status code. Use BaseModule.status instead.'
            )

        status_dict = {'status': status, 'msg': msg, 'args': {}}
        if args:
            if 'code' in args.keys():
                args[
                    'code'
                ] = f'{self.package_name.upper()}_{self.module_name.upper()}_{args["code"]}'
            status_dict['args'] = args
        return MethodException(status_dict)

    def __getattribute__(self, attr):
        # [DOC] Module is not yet initialised, skip to return exact attr
        try:
            object.__getattribute__(self, 'methods')
        except AttributeError:
            return object.__getattribute__(self, attr)
        # [DOC] Module is initialised attempt to check for methods
        if attr in object.__getattribute__(self, 'methods').keys():
            return object.__getattribute__(self, 'methods')[attr]

        if attr.startswith('_method_'):
            return object.__getattribute__(self, attr.replace('_method_', ''))

        return object.__getattribute__(self, attr)

    async def pre_read(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_read(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def read(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> RESULTS:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'read\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        if Event.PRE not in skip_events:
            pre_read = await self.pre_read(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_read

            # [DOC] Check if __results are passed in payload
            if '__results' in payload.keys():
                return payload['__results']

        results = await Data.read(
            env=env,
            collection_name=self.collection,
            attrs=self.attrs,
            query=query,
            skip_extn='$extn' in query or Event.EXTN in skip_events,
        )

        if Event.ON not in skip_events:
            on_read = await self.on_read(
                results=results,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                payload=payload,
            )
            results, skip_events, env, query, doc, payload = on_read
            # [DOC] if $attrs query arg is present return only required keys.
            if '$attrs' in query:
                query['$attrs'].insert(0, '_id')
                for i in range(len(results['docs'])):
                    results['docs'][i] = {
                            attr: results['docs'][i][attr]
                            for attr in query['$attrs']
                            if attr in results['docs'][i]
                        }

        return self.status(
            status=200, msg=f'Found {results["count"]} docs.', args=results
        )

    async def pre_read_diff(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_read_diff(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def read_diff(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> RESULTS:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'read_diff\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        if Event.PRE not in skip_events:
            pre_read_diff = await self.pre_read_diff(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_read_diff

            # [DOC] Check if __results are passed in payload
            if '__results' in payload.keys():
                return payload['__results']

        results = await Data.read(
            env=env,
            collection_name=f'{self.collection}_diff',
            attrs=diff_attrs,
            query=query,
            skip_extn='$extn' in query or Event.EXTN in skip_events,
        )

        if Event.ON not in skip_events:
            on_read_diff = await self.on_read_diff(
                results=results,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                payload=payload,
            )
            results, skip_events, env, query, doc, payload = on_read_diff

        return self.status(
            status=200, msg=f'Found {results["count"]} docs.', args=results
        )

    async def pre_watch(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_watch(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def watch(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> AsyncGenerator[Dict, Dict]:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'watch\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        if Event.PRE not in skip_events:
            pre_watch = await self.pre_watch(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_watch

        logger.debug('Preparing async loop at BaseModule')
        self.collection = cast(str, self.collection)
        async for results in Data.watch(
            env=env,
            collection_name=self.collection,
            attrs=self.attrs,
            query=query,
            skip_extn='$extn' in query or Event.EXTN in skip_events,
        ):
            logger.debug(f'Received watch results at BaseModule: {results}')

            if 'stream' in results.keys():
                yield self.status(
                    status=200, msg=f'Detected {results["count"]} docs.', args=results
                )
                continue

            if Event.ON not in skip_events:
                on_watch = await self.on_watch(
                    results=results,
                    skip_events=skip_events,
                    env=env,
                    query=query,
                    doc=doc,
                    payload=payload,
                )
                results, skip_events, env, query, doc, payload = on_watch

                # [DOC] if $attrs query arg is present return only required keys.
                if '$attrs' in query:
                    query['$attrs'].insert(0, '_id')
                    for i in range(len(results['docs'])):
                        results['docs'][i] = {
                                attr: results['docs'][i][attr]
                                for attr in query['$attrs']
                                if attr in results['docs'][i]
                            }

            yield self.status(
                status=200, msg=f'Detected {results["count"]} docs.', args=results
            )

        logger.debug('Generator ended at BaseModule.')

    async def pre_create(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_create(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def create(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> RESULTS:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'create\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        # [DOC] Check for __create_draft
        if '__create_draft' in query:
            # [DOC] Check for conflicts
            if len(doc.keys()):
                raise self.exception(
                    status=400,
                    msg='Can\'t use \'__create_draft\' with attrs provided in \'doc\'',
                    args={'code': 'INVALID_CREATE_DRAFT_DOC'},
                )

            # [DOC] Load __create_draft and use it as doc
            draft_results = await Data.read(
                env=env,
                collection_name=self.collection,
                attrs=self.attrs,
                query=Query(
                    [{'_id': query['__create_draft'][0], '__create_draft': True}]
                ),
                skip_process=True,
            )

            if not draft_results['count']:
                raise self.exception(
                    status=400,
                    msg='Invalid \'__create_draft\' doc.',
                    args={'code': 'INVALID_CREATE_DRAFT'},
                )

            # [TODO] Check copy.deepcopy need
            doc = draft_results['docs'][0]
            # [DOC] Delete special attrs
            del doc['_id']

        if Event.PRE not in skip_events:
            pre_create = await self.pre_create(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_create

            # [DOC] Check if __results are passed in payload
            if '__results' in payload.keys():
                return payload['__results']

        # [DOC] Expant dot-notated keys onto dicts
        doc = _expand_attr(doc=doc)
        # [DOC] Deleted all extra doc args
        doc = {
            attr: doc[attr]
            for attr in ['_id', '__create_draft', '__update_draft', *self.attrs.keys()]
            if attr in doc.keys() and doc[attr] != None
        }
        # [DOC] Append host_add, user_agent, create_time, if it's present in attrs.
        if (
            'user' in self.attrs.keys()
            and 'host_add' not in doc.keys()
            and env['session']
            and Event.ARGS not in skip_events
        ):
            doc['user'] = env['session']['user']['_id']
        if 'create_time' in self.attrs.keys():
            doc['create_time'] = datetime.datetime.utcnow().isoformat()
        if 'host_add' in self.attrs.keys() and 'host_add' not in doc.keys():
            doc['host_add'] = env['REMOTE_ADDR']
        if 'user_agent' in self.attrs.keys() and 'user_agent' not in doc.keys():
            doc['user_agent'] = env['HTTP_USER_AGENT']
        if Event.ARGS not in skip_events:
            mode: Literal['create', 'create_draft'] = 'create'
            if '__create_draft' in doc.keys() and doc['__create_draft'] == True:
                if not self.create_draft:
                    raise self.exception(
                        status=400,
                        msg=f'Module \'{self.package_name.upper()}_{self.module_name.upper()}\' doesn\'t support \'create_draft\'',
                        args={'code': 'NO_CREATE_DRAFT'},
                    )
                if type(self.create_draft) == ATTR:
                    self.create_draft = cast(ATTR, self.create_draft)
                    # [DOC] Attr Type TYPE create_draft, call the funcion and catch InvalidAttrException
                    try:
                        func_params = {
                            'mode': 'create',
                            'attr_name': 'create_draft',
                            'attr_type': self.create_draft,
                            'attr_val': None,
                            'skip_events': skip_events,
                            'env': env,
                            'query': query,
                            'doc': doc,
                            'scope': doc,
                        }
                        await self.create_draft._args['func'](
                            **{
                                param: func_params[param]
                                for param in inspect.signature(
                                    self.create_draft._args['func']
                                ).parameters
                            }
                        )
                    except:
                        raise self.exception(
                            status=400,
                            msg=f'Module \'{self.package_name.upper()}_{self.module_name.upper()}\' \'create_draft\' failed.',
                            args={'code': 'NO_CREATE_DRAFT_CONDITION'},
                        )

                mode = 'create_draft'

            elif '__update_draft' in doc.keys() and (
                type(doc['__update_draft']) == ObjectId
                or (
                    type(doc['__update_draft']) == str
                    and re.match(r'^[0-9a-fA-F]{24}$', doc['__update_draft'])
                )
            ):
                doc['__update_draft'] = ObjectId(doc['__update_draft'])
                if not self.update_draft:
                    raise self.exception(
                        status=400,
                        msg=f'Module \'{self.package_name.upper()}_{self.module_name.upper()}\' doesn\'t support \'update_draft\'',
                        args={'code': 'NO_UPDATE_DRAFT'},
                    )
                if type(self.update_draft) == ATTR:
                    self.update_draft = cast(ATTR, self.update_draft)
                    # [DOC] Attr Type TYPE update_draft, call the funcion and catch InvalidAttrException
                    try:
                        func_params = {
                            'mode':'create',
                            'attr_name':'update_draft',
                            'attr_type':self.update_draft,
                            'attr_val':None,
                            'skip_events':skip_events,
                            'env':env,
                            'query':query,
                            'doc':doc,
                            'scope':doc,
                        }
                        await self.update_draft._args['func'](
                            **{
                                param: func_params[param]
                                for param in inspect.signature(
                                    self.update_draft._args['func']
                                ).parameters
                            }
                        )
                    except:
                        raise self.exception(
                            status=400,
                            msg=f'Module \'{self.package_name.upper()}_{self.module_name.upper()}\' \'update_draft\' failed.',
                            args={'code': 'NO_UPDATE_DRAFT_CONDITION'},
                        )

                mode = 'create_draft'
            # [DOC] Check presence and validate all attrs in doc args
            try:
                await validate_doc(
                    mode=mode,
                    doc=doc,
                    attrs=self.attrs,
                    skip_events=skip_events,
                    env=env,
                    query=query,
                )
            except MissingAttrException as e:
                raise self.exception(
                    status=400,
                    msg=f'{str(e)} for \'create\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
                    args={'code': 'MISSING_ATTR'},
                )
            except InvalidAttrException as e:
                raise self.exception(
                    status=400,
                    msg=f'{str(e)} for \'create\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
                    args={'code': 'INVALID_ATTR'},
                )
            except ConvertAttrException as e:
                raise self.exception(
                    status=400,
                    msg=f'{str(e)} for \'create\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
                    args={'code': 'CONVERT_INVALID_ATTR'},
                )
            # [DOC] Check unique_attrs
            if self.unique_attrs:
                unique_attrs_query: List[Any] = [[]]
                for attr in self.unique_attrs:
                    if type(attr) == str:
                        attr = cast(str, attr)
                        unique_attrs_query[0].append({attr: doc[attr]})
                    elif type(attr) == tuple:
                        unique_attrs_query[0].append(
                            {child_attr: doc[child_attr] for child_attr in attr}
                        )
                    # [TODO] Implement use of single-item dict with LITERAL Attr Type for dynamic unique check based on doc value
                unique_attrs_query.append({'$limit': 1})
                unique_attrs_query = cast(NAWAH_QUERY, unique_attrs_query)
                unique_results = await self.read(
                    skip_events=[Event.PERM], env=env, query=unique_attrs_query
                )
                if unique_results['args']['count']:
                    unique_attrs_str = ', '.join(
                        map(
                            lambda _: ('(' + ', '.join(_) + ')') if type(_) == tuple else _,  # type: ignore
                            self.unique_attrs,
                        )
                    )
                    raise self.exception(
                        status=400,
                        msg=f'A doc with the same \'{unique_attrs_str}\' already exists.',
                        args={'code': 'DUPLICATE_DOC'},
                    )
        # [DOC] Execute Data driver create
        results = await Data.create(
            env=env, collection_name=self.collection, attrs=self.attrs, doc=doc
        )

        # [DOC] Check for __create_draft and delete it
        if '__create_draft' in query:
            delete_draft_results = await Data.delete(
                env=env,
                collection_name=self.collection,
                attrs=self.attrs,
                docs=[ObjectId(query['__create_draft'][0])],
                strategy=DELETE_STRATEGY.FORCE_SYS,
            )
            if delete_draft_results['count'] != 1:
                logger.error(
                    f'Failed to delete \'__create_draft\'. Results: {delete_draft_results}'
                )

        if Event.ON not in skip_events:
            on_create = await self.on_create(
                results=results,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                payload=payload,
            )
            results, skip_events, env, query, doc, payload = on_create

        # [DOC] create soft action is to only return the new created doc _id.
        if Event.SOFT in skip_events:
            read_results = await self.read(
                skip_events=[Event.PERM], env=env, query=[[{'_id': results['docs'][0]}]]
            )
            results = read_results['args']

        return self.status(
            status=200, msg=f'Created {results["count"]} docs.', args=results
        )

    async def pre_update(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_update(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def update(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> RESULTS:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'update\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        update_draft = None
        # [DOC] Check for __update_draft
        if '__update_draft' in query:
            update_draft = query['__update_draft'][0]
            # [DOC] Check for conflicts
            if len(doc.keys()):
                raise self.exception(
                    status=400,
                    msg='Can\'t use \'__update_draft\' with attrs provided in \'doc\'',
                    args={'code': 'INVALID_UPDATE_DRAFT_DOC'},
                )

            # [DOC] Load __create_draft and use it as doc
            draft_results = await Data.read(
                env=env,
                collection_name=self.collection,
                attrs=self.attrs,
                query=Query(
                    [
                        {
                            '_id': query['__update_draft'][0],
                            '__update_draft': {'$ne': False},
                        }
                    ]
                ),
                skip_process=True,
            )

            if not draft_results['count']:
                raise self.exception(
                    status=400,
                    msg='Invalid \'__update_draft\' doc.',
                    args={'code': 'INVALID_UPDATE_DRAFT'},
                )

            # [DOC] Update query per the value of doc.__update_draft
            del query['__update_draft'][0]
            query.append({'_id': draft_results['docs'][0]['__update_draft']})
            # [TODO] Check copy.deepcopy need
            doc = draft_results['docs'][0]
            # [DOC] Delete special attrs
            del doc['_id']

        if Event.PRE not in skip_events:
            pre_update = await self.pre_update(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_update

            # [DOC] Check if __results are passed in payload
            if '__results' in payload.keys():
                return payload['__results']

        # [DOC] Check presence and validate all attrs in doc args
        try:
            await validate_doc(
                mode='update',
                doc=doc,
                attrs=self.attrs,
                skip_events=skip_events,
                env=env,
                query=query,
            )
        except MissingAttrException as e:
            raise self.exception(
                status=400,
                msg=f'{str(e)} for \'update\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
                args={'code': 'MISSING_ATTR'},
            )
        except InvalidAttrException as e:
            raise self.exception(
                status=400,
                msg=f'{str(e)} for \'update\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
                args={'code': 'INVALID_ATTR'},
            )
        except ConvertAttrException as e:
            raise self.exception(
                status=400,
                msg=f'{str(e)} for \'update\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
                args={'code': 'CONVERT_INVALID_ATTR'},
            )
        # [DOC] Delete all attrs not belonging to the module
        shadow_doc = {}
        for attr_name in ['_id', *doc.keys()]:
            attr_root = attr_name.split('.')[0].split(':')[0]
            # [DOC] Check top-level attr belong to module
            if attr_root not in self.attrs.keys():
                continue
            # [DOC] Check attr is valid Doc Oper
            if (
                type(doc[attr_name]) == dict
                and doc[attr_name].keys()
                and list(doc[attr_name].keys())[0][0] == '$'
                and doc[attr_name][list(doc[attr_name].keys())[0]] == None
            ):
                continue
            # [DOC] Add non-None attrs to shadow_doc
            if doc[attr_name] != None:
                shadow_doc[attr_name] = doc[attr_name]

        # [DOC] Replace all NAWAH_VALUES.NONE_VALUE with None
        _set_none_value(doc=shadow_doc)
        doc = shadow_doc

        # [DOC] Check if there is anything yet to update
        if not len(doc.keys()):
            return self.status(status=200, msg='Nothing to update.', args={})
        # [DOC] Find which docs are to be updated
        docs_results = await Data.read(
            env=env,
            collection_name=self.collection,
            attrs=self.attrs,
            query=query,
            skip_process=True,
        )
        # [DOC] Check unique_attrs
        if self.unique_attrs:
            # [DOC] If any of the unique_attrs is present in doc, and docs_results is > 1, we have duplication
            if len(docs_results['docs']) > 1:
                unique_attrs_check = True
                for attr in self.unique_attrs:
                    if type(attr) == str and attr in doc.keys():
                        unique_attrs_check = False
                        break
                    elif type(attr) == tuple:
                        for child_attr in attr:
                            if not unique_attrs_check:
                                break
                            if child_attr in doc.keys():
                                unique_attrs_check = False
                                break

                if not unique_attrs_check:
                    raise self.exception(
                        status=400,
                        msg='Update call query has more than one doc as results. This would result in duplication.',
                        args={'code': 'MULTI_DUPLICATE'},
                    )

            # [DOC] Check if any of the unique_attrs are present in doc
            # [TODO] Replace with any
            if sum(1 for attr in doc.keys() if attr in self.unique_attrs) > 0:
                # [DOC] Check if the doc would result in duplication after update
                unique_attrs_query: List[Any] = [[]]
                for attr in self.unique_attrs:
                    if type(attr) == str:
                        attr = cast(str, attr)
                        if attr in doc.keys():
                            unique_attrs_query[0].append({attr: doc[attr]})
                    elif type(attr) == tuple:
                        unique_attrs_query[0].append(
                            {
                                child_attr: doc[child_attr]
                                for child_attr in attr
                                if attr in doc.keys()
                            }
                        )
                unique_attrs_query.append(
                    {'_id': {'$nin': [doc['_id'] for doc in docs_results['docs']]}}
                )
                unique_attrs_query.append({'$limit': 1})
                unique_attrs_query = cast(NAWAH_QUERY, unique_attrs_query)
                unique_results = await self.read(
                    skip_events=[Event.PERM], env=env, query=unique_attrs_query
                )
                if unique_results['args']['count']:
                    unique_attrs_str = ', '.join(
                        map(
                            lambda _: ('(' + ', '.join(_) + ')') if type(_) == tuple else _,  # type: ignore
                            self.unique_attrs,
                        )
                    )
                    raise self.exception(
                        status=400,
                        msg=f'A doc with the same \'{unique_attrs_str}\' already exists.',
                        args={'code': 'DUPLICATE_DOC'},
                    )
        create_diff = False
        # [DOC] If module has diff enabled, and Event DIFF not skipped:
        if self.diff and Event.DIFF not in skip_events:
            create_diff = True
            if type(self.diff) == ATTR:
                # [DOC] # [DOC] Attr Type TYPE diff, call the funcion and catch InvalidAttrException
                self.diff = cast(ATTR, self.diff)
                try:
                    func_params = {
                        'mode':'create',
                        'attr_name':'diff',
                        'attr_type':self.diff,
                        'attr_val':None,
                        'skip_events':skip_events,
                        'env':env,
                        'query':query,
                        'doc':doc,
                        'scope':doc,
                    }
                    await self.diff._args['func'](
                        **{
                            param: func_params[param]
                            for param in inspect.signature(
                                self.diff._args['func']
                            ).parameters
                        }
                    )

                except:
                    create_diff = False
                    logger.debug(f'Skipped Diff Workflow due to failed condition.')

        else:
            logger.debug(
                f'Skipped Diff Workflow due to: {self.diff}, {Event.DIFF not in skip_events}'
            )

        results = await Data.update(
            env=env,
            collection_name=self.collection,
            attrs=self.attrs,
            docs=[doc['_id'] for doc in docs_results['docs']],
            doc=doc,
        )

        if results['count'] and create_diff:
            for doc_result in docs_results['docs']:
                diff_results = await Data.create(
                    env=env,
                    collection_name=f'{self.collection}_diff',
                    attrs=self.attrs,
                    doc={
                        'user': env['session']['user']['_id'],
                        'doc': doc_result['_id'],
                        'attrs': {
                            attr.split('.')[0]: doc_result[attr.split('.')[0]]
                            for attr in doc.keys()
                        },
                    },
                )
            if not diff_results['count']:
                logger.error(f'Failed to create Diff doc. Results: {diff_results}')

        # [DOC] Check for update_draft and delete it
        if update_draft:
            delete_draft_results = await Data.delete(
                env=env,
                collection_name=self.collection,
                attrs=self.attrs,
                docs=[ObjectId(update_draft)],
                strategy=DELETE_STRATEGY.FORCE_SYS,
            )
            if delete_draft_results['count'] != 1:
                logger.error(
                    f'Failed to delete \'__update_draft\'. Results: {delete_draft_results}'
                )

        if Event.ON not in skip_events:
            on_update = await self.on_update(
                results=results,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                payload=payload,
            )
            results, skip_events, env, query, doc, payload = on_update

        return self.status(
            status=200, msg=f'Updated {results["count"]} docs.', args=results
        )

    async def pre_delete(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_delete(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def delete(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> RESULTS:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'delete\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        if Event.PRE not in skip_events:
            pre_delete = await self.pre_delete(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_delete

            # [DOC] Check if __results are passed in payload
            if '__results' in payload.keys():
                return payload['__results']

        # [TODO]: confirm all extns are not linked.
        # [DOC] Pick delete strategy based on skip_events
        strategy = DELETE_STRATEGY.SOFT_SKIP_SYS
        if Event.SOFT not in skip_events and Event.SYS_DOCS in skip_events:
            strategy = DELETE_STRATEGY.SOFT_SYS
        elif Event.SOFT in skip_events:
            if Event.SYS_DOCS not in skip_events:
                strategy = DELETE_STRATEGY.FORCE_SKIP_SYS
            else:
                strategy = DELETE_STRATEGY.FORCE_SYS

        docs_results = results = await Data.read(
            env=env,
            collection_name=self.collection,
            attrs=self.attrs,
            query=query,
            skip_process=True,
        )
        results = await Data.delete(
            env=env,
            collection_name=self.collection,
            attrs=self.attrs,
            docs=[doc['_id'] for doc in docs_results['docs']],
            strategy=strategy,
        )
        if Event.ON not in skip_events:
            on_delete = await self.on_delete(
                results=results,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                payload=payload,
            )
            results, skip_events, env, query, doc, payload = on_delete

        return self.status(
            status=200, msg=f'Deleted {results["count"]} docs.', args=results
        )

    async def pre_create_file(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_create_file(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def create_file(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> RESULTS:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'create_file\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        if Event.PRE not in skip_events:
            pre_create_file = await self.pre_create_file(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_create_file

            # [DOC] Check if __results are passed in payload
            if '__results' in payload.keys():
                return payload['__results']

        # [TODO] Allow use dot-notated attr path in attr query attr
        if (
            query['attr'][0] not in self.attrs.keys()
            or type(self.attrs[query['attr'][0]]._type) != 'LIST'
            or not self.attrs[query['attr'][0]]._args['list'][0]._type != 'FILE'
        ):
            raise self.exception(
                status=400, msg='Attr is invalid.', args={'code': 'INVALID_ATTR'}
            )

        update_results = await self.update(
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': query['_id'][0]}],
            doc={query['attr'][0]: {'$append': doc['file']}},
        )
        results = update_results['args']

        if Event.ON not in skip_events:
            on_create_file = await self.on_create_file(
                results=results,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                payload=payload,
            )
            results, skip_events, env, query, doc, payload = on_create_file

        return self.status(
            status=200, msg=f'Updated {results["count"]} docs.', args=results
        )

    async def pre_delete_file(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_delete_file(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def delete_file(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> RESULTS:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'delete_file\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        if Event.PRE not in skip_events:
            pre_delete_file = await self.pre_delete_file(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_delete_file

            # [DOC] Check if __results are passed in payload
            if '__results' in payload.keys():
                return payload['__results']

        # [TODO] Allow use dot-notated attr path in attr query attr
        if (
            query['attr'][0] not in self.attrs.keys()
            or type(self.attrs[query['attr'][0]]._type) != 'LIST'
            or not self.attrs[query['attr'][0]]._args['list'][0]._type != 'FILE'
        ):
            raise self.exception(
                status=400, msg='Attr is invalid.', args={'code': 'INVALID_ATTR'}
            )

        read_results = await self.read(
            skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
        )
        if not read_results['args']['count']:
            raise self.exception(
                status=400, msg='Doc is invalid.', args={'code': 'INVALID_DOC'}
            )
        doc = read_results['args']['docs'][0]

        if query['attr'][0] not in doc or not doc[query['attr'][0]]:
            raise self.exception(
                status=400,
                msg='Doc attr is invalid.',
                args={'code': 'INVALID_DOC_ATTR'},
            )

        if query['index'][0] not in range(len(doc[query['attr'][0]])):
            raise self.exception(
                status=400, msg='Index is invalid.', args={'code': 'INVALID_INDEX'}
            )

        if (
            type(doc[query['attr'][0]][query['index'][0]]) != dict
            or 'name' not in doc[query['attr'][0]][query['index'][0]].keys()
        ):
            raise self.exception(
                status=400,
                msg='Index value is invalid.',
                args={'code': 'INVALID_INDEX_VALUE'},
            )

        if doc[query['attr'][0]][query['index'][0]]['name'] != query['name'][0]:  # type: ignore
            raise self.exception(
                status=400,
                msg='File name in query doesn\'t match value.',
                args={'code': 'FILE_NAME_MISMATCH'},
            )

        update_results = await self.update(
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': query['_id'][0]}],
            doc={
                query['attr'][0]: {
                    '$del_val': [doc[query['attr'][0]][query['index'][0]]]
                }
            },
        )
        results = update_results['args']

        if Event.ON not in skip_events:
            on_delete_file = await self.on_delete_file(
                results=results,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                payload=payload,
            )
            results, skip_events, env, query, doc, payload = on_delete_file

        return results

    async def pre_retrieve_file(
        self,
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> PRE_HANDLER_RETURN:
        return (skip_events, env, query, doc, payload)

    async def on_retrieve_file(
        self,
        results: Dict[str, Any],
        skip_events: NAWAH_EVENTS,
        env: NAWAH_ENV,
        query: Query,
        doc: NAWAH_DOC,
        payload: Dict[str, Any],
    ) -> ON_HANDLER_RETURN:
        return (results, skip_events, env, query, doc, payload)

    async def retrieve_file(
        self,
        skip_events: NAWAH_EVENTS = [],
        env: NAWAH_ENV = {},
        query: Union[NAWAH_QUERY, Query] = [],
        doc: NAWAH_DOC = {},
    ) -> RESULTS:
        if not self.collection:
            raise self.exception(
                status=400,
                msg='Utility module can\'t call \'retrieve_file\' method.',
                args={'code': 'INVALID_CALL'},
            )

        payload: Dict[str, Any] = {}

        query = cast(Query, query)

        if Event.PRE not in skip_events:
            pre_retrieve_file = await self.pre_retrieve_file(
                skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
            )
            skip_events, env, query, doc, payload = pre_retrieve_file

            # [DOC] Check if __results are passed in payload
            if '__results' in payload.keys():
                return payload['__results']

        attr_name = query['attr'][0]
        filename = query['filename'][0]
        if 'thumb' in query:
            thumb_dims: Optional[List[int]] = [
                int(dim) for dim in query['thumb'][0].split('x')
            ]
        else:
            thumb_dims = None

        read_results = await self.read(
            skip_events=[Event.PERM, Event.CACHE] + skip_events,
            env=env,
            query=[{'_id': query['_id'][0]}],
        )
        if not read_results['args']['count']:
            raise self.exception(
                status=400,
                msg='File not found.',
                args={'code': 'NOT_FOUND', 'return': 'json'},
            )
        attr: Union[List[Dict[str, Any]], Dict[str, Any]]
        doc = read_results['args']['docs'][0]
        try:
            attr_path = attr_name.split('.')
            attr = doc
            for path in attr_path:
                attr = doc[path]
        except:
            raise self.exception(
                status=404,
                msg='File not found.',
                args={'code': 'NOT_FOUND', 'return': 'json'},
            )

        retrieved_file = None

        if type(attr) == list:
            attr = cast(List[Dict[str, Any]], attr)
            for item in attr:
                if item['name'] == filename:
                    retrieved_file = item
                    break
        elif type(attr) == dict:
            attr = cast(Dict[str, Any], attr)
            if attr['name'] == filename:
                retrieved_file = attr

        if not retrieved_file:
            # [DOC] No filename match
            raise self.exception(
                status=404,
                msg='File not found.',
                args={'code': 'NOT_FOUND', 'return': 'json'},
            )

        # [DOC] filematch!
        results: Dict[str, Any]
        results = {
            'docs': [
                    {
                        '_id': query['_id'][0],
                        'name': retrieved_file['name'],
                        'type': retrieved_file['type'],
                        'lastModified': retrieved_file['lastModified'],
                        'size': retrieved_file['size'],
                        'content': retrieved_file['content'],
                    }
            ]
        }

        if thumb_dims:
            if retrieved_file['type'].split('/')[0] != 'image':
                raise self.exception(
                    status=400,
                    msg='File is not of type image to create thumbnail for.',
                    args={'code': 'NOT_IMAGE', 'return': 'json'},
                )
            try:
                image = Image.open(io.BytesIO(retrieved_file['content']))
                image.thumbnail(thumb_dims)
                stream = io.BytesIO()
                image.save(stream, format=image.format)
                stream.seek(0)
                results['docs'][0]['content'] = stream.read()
            except:
                pass

        if Event.ON not in skip_events:
            on_retrieve_file = await self.on_retrieve_file(
                results=results,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                payload=payload,
            )
            results, skip_events, env, query, doc, payload = on_retrieve_file

        results['return'] = 'file'
        return self.status(status=200, msg='File attached to response.', args=results)



import copy
from typing import (TYPE_CHECKING, Any, Dict, List, Literal, Optional,
                    TypedDict, Union, cast)

from ._attr import SPECIAL_ATTRS
from ._exceptions import InvalidQueryArgException, UnknownQueryArgException

if TYPE_CHECKING:
    from ._types import NAWAH_QUERY, NAWAH_QUERY_SPECIAL


QUERY_INDEX_RECORD = TypedDict(
    'QUERY_INDEX_RECORD', {'oper': str, 'path': str, 'val': Any}
)


class Query(list):
    _query: 'NAWAH_QUERY'
    _special: 'NAWAH_QUERY_SPECIAL'
    _index: Dict[str, List[QUERY_INDEX_RECORD]]

    def __init__(self, query: Union['NAWAH_QUERY', 'Query']):
        self._query = query
        if type(self._query) == Query:
            query = cast(Query, query)
            self._query = query._query + [query._special]
        self._special = {}
        self._index: Dict[str, List[QUERY_INDEX_RECORD]] = {}
        self._create_index(self._query)
        super().__init__(self._query)

    def __repr__(self):
        return str(self._query + [self._special])

    def _create_index(self, query: 'NAWAH_QUERY', path=[]):
        if not path:
            self._index: Dict[str, List[QUERY_INDEX_RECORD]] = {}
        for i in range(len(query)):
            if type(query[i]) == dict:
                del_attrs = []
                for attr in query[i].keys():
                    if attr in SPECIAL_ATTRS.__args__:  # type: ignore
                        self._special[attr] = query[i][attr]  # type: ignore
                        del_attrs.append(attr)
                    elif attr.startswith('__or'):
                        self._create_index(query[i][attr], path=path + [i, attr])  # type: ignore
                    else:
                        if (
                            type(query[i][attr]) == dict  # type: ignore
                            and len(query[i][attr].keys()) == 1  # type: ignore
                            and list(query[i][attr].keys())[0][0] == '$'  # type: ignore
                        ):
                            attr_oper = list(query[i][attr].keys())[0]  # type: ignore
                        else:
                            attr_oper = '$eq'
                        if attr not in self._index.keys():
                            self._index[attr] = []
                        if isinstance(query[i][attr], Dict) and '_id' in query[i][attr]:  # type: ignore
                            query[i][attr] = query[i][attr]['_id']  # type: ignore
                        Query.validate_arg(arg_name=attr, arg_oper=attr_oper, arg_val=query[i][attr])  # type: ignore
                        self._index[attr].append(
                            {
                                'oper': attr_oper,
                                'path': path + [i],
                                'val': query[i][attr],  # type: ignore
                            }
                        )
                for attr in del_attrs:
                    del query[i][attr]  # type: ignore
            elif type(query[i]) == list:
                self._create_index(query[i], path=path + [i])  # type: ignore
        if not path:
            self._query = self._sanitise_query()

    def _sanitise_query(self, query: 'NAWAH_QUERY' = None):
        if query == None:
            query = self._query
        query = cast('NAWAH_QUERY', query)
        query_shadow = []
        for step in query:
            if type(step) == dict:
                for attr in step.keys():
                    if attr.startswith('__or'):
                        step[attr] = self._sanitise_query(step[attr])  # type: ignore
                        if len(step[attr]):  # type: ignore
                            query_shadow.append(step)
                            break
                    elif attr[0] != '$':
                        query_shadow.append(step)
                        break
            elif type(step) == list:
                step = self._sanitise_query(step)  # type: ignore
                if len(step):
                    query_shadow.append(step)
        return query_shadow

    def __deepcopy__(self, memo):
        return Query(copy.deepcopy(self._query + [self._special]))

    def append(self, obj: Any):
        self._query.append(obj)
        self._create_index(self._query)
        super().__init__(self._query)

    def __contains__(self, attr: str):  # type: ignore
        if attr in SPECIAL_ATTRS.__args__:  # type: ignore
            return attr in self._special.keys()
        else:
            if ':' in attr:
                attr_index, attr_oper = attr.split(':')
            else:
                attr_index = attr
                attr += ':$eq'
                attr_oper = '$eq'

            if attr_index in self._index.keys():
                for val in self._index[attr_index]:
                    if val['oper'] == attr_oper or attr_oper == '*':
                        return True
            return False

    def __getitem__(self, attr: Union[SPECIAL_ATTRS, str]):  # type: ignore
        if attr in SPECIAL_ATTRS.__args__:  # type: ignore
            return self._special[attr]  # type: ignore
        else:
            attrs = []
            vals = []
            paths: List[List[int]] = []
            indexes = []
            attr_filter: Optional[str] = None
            oper_filter: Optional[str] = None

            if attr.split(':')[0] != '*':
                attr_filter = attr.split(':')[0]

            if ':' not in attr:
                oper_filter = '$eq'
                attr += ':$eq'
            elif ':*' not in attr:
                oper_filter = attr.split(':')[1]

            for index_attr in self._index.keys():
                if attr_filter and index_attr != attr_filter:
                    continue
                i = 0
                for val in self._index[index_attr]:
                    if not oper_filter or (oper_filter and val['oper'] == oper_filter):
                        attrs.append(index_attr)
                        # [TODO] Simplify this condition by enforcing Query Args with $eq Query Oper are always stripped down to value
                        if not oper_filter or (
                            oper_filter == '$eq'
                            and (
                                type(val['val']) != dict
                                or '$eq' not in val['val'].keys()
                            )
                        ):
                            vals.append(val['val'])
                        else:
                            vals.append(val['val'][oper_filter])
                        paths.append(val['path'])
                        indexes.append(i)
                        i += 1
            return QueryAttrList(self, attrs, paths, indexes, vals)

    def __setitem__(self, attr: str, val: Any):  # type: ignore
        if attr[0] != '$':
            raise Exception('Non-special attrs can only be updated by attr index.')
        self._special[attr] = val  # type: ignore

    def __delitem__(self, attr: str):  # type: ignore
        if attr[0] != '$':
            raise Exception('Non-special attrs can only be deleted by attr index.')
        del self._special[attr]  # type: ignore

    @classmethod
    def validate_arg(cls, *, arg_name, arg_oper, arg_val):
        if arg_oper in ['$ne', '$eq']:
            return
        elif arg_oper in ['$gt', '$gte', '$lt', '$lte']:
            if type(arg_val[arg_oper]) not in [str, int, float]:
                raise InvalidQueryArgException(
                    arg_name=arg_name,
                    arg_oper=arg_oper,
                    arg_type=[str, int, float],
                    arg_val=arg_val[arg_oper],
                )
        elif arg_oper == '$bet':
            if (
                type(arg_val[arg_oper]) != list
                or len(arg_val[arg_oper]) != 2
                or type(arg_val[arg_oper][0]) not in [str, int, float]
                or type(arg_val[arg_oper][1]) not in [str, int, float]
            ):
                raise InvalidQueryArgException(
                    arg_name=arg_name,
                    arg_oper=arg_oper,
                    arg_type=list,
                    arg_val=arg_val[arg_oper],
                )
        elif arg_oper in ['$all', '$in', '$nin']:
            if type(arg_val[arg_oper]) != list:
                # [TODO] Validate why this was added
                # or not len(arg_val[arg_oper]):
                raise InvalidQueryArgException(
                    arg_name=arg_name,
                    arg_oper=arg_oper,
                    arg_type=list,
                    arg_val=arg_val[arg_oper],
                )
        elif arg_oper == '$regex':
            if type(arg_val[arg_oper]) != str:
                raise InvalidQueryArgException(
                    arg_name=arg_name,
                    arg_oper=arg_oper,
                    arg_type=str,
                    arg_val=arg_val[arg_oper],
                )
        else:
            raise UnknownQueryArgException(arg_name=arg_name, arg_oper=arg_oper)


class QueryAttrList(list):
    def __init__(
        self,
        query: Query,
        attrs: List[str],
        paths: List[List[int]],
        indexes: List[int],
        vals: List[Any],
    ):
        self._query = query
        self._attrs = attrs
        self._paths = paths
        self._indexes = indexes
        self._vals = vals
        super().__init__(vals)

    def __setitem__(self, item: Union[Literal['*'], int], val: Any):  # type: ignore
        if item == '*':
            for i in range(len(self._vals)):
                self.__setitem__(i, val)
        else:
            instance_attr = self._query._query
            for path_part in self._paths[item]:
                instance_attr = instance_attr[path_part]  # type: ignore
            instance_attr[self._attrs[item].split(':')[0]] = val  # type: ignore
            self._query._create_index(self._query._query)

    def __delitem__(self, item: Union[Literal['*'], int]):  # type: ignore
        if item == '*':
            for i in range(len(self._vals)):
                self.__delitem__(i)
        else:
            instance_attr = self._query._query
            for path_part in self._paths[item]:
                instance_attr = instance_attr[path_part]  # type: ignore
            del instance_attr[self._attrs[item].split(':')[0]]  # type: ignore
            self._query._create_index(self._query._query)

    def replace_attr(self, item: Union[Literal['*'], int], new_attr: str):
        if item == '*':
            for i in range(len(self._vals)):
                self.replace_attr(i, new_attr)
        else:
            instance_attr = self._query._query
            for path_part in self._paths[item]:
                instance_attr = instance_attr[path_part]  # type: ignore
            # [DOC] Set new attr
            instance_attr[new_attr] = instance_attr[self._attrs[item].split(':')[0]]  # type: ignore
            # [DOC] Delete old attr
            del instance_attr[self._attrs[item].split(':')[0]]  # type: ignore
            # [DOC] Update index
            self._query._create_index(self._query._query)



import datetime
import logging
import secrets

from passlib.hash import pbkdf2_sha512

from nawah.base_module import BaseModule
from nawah.classes import ATTR, EXTN, METHOD, PERM
from nawah.config import Config
from nawah.enums import Event
from nawah.registry import Registry

logger = logging.getLogger('nawah')


class Session(BaseModule):
    '''`Session` module provides data type and controller for sessions in Nawah eco-system. CRUD methods of the module are supposed to used for internal calls only, while methods `auth`, `reauth`, and `signout` are available for use by API as well as internal calls when needed.'''

    collection = 'sessions'
    attrs = {
        'user': ATTR.ID(desc='`_id` of `User` doc the doc belongs to.'),
        'groups': ATTR.LIST(
            desc='List of `_id` for every group the session is authenticated against. This attr is set by `auth` method when called with `groups` Doc Arg for Controller Auth Sequence.',
            list=[
                ATTR.ID(desc='`_id` of Group doc the session is authenticated against.')
            ],
        ),
        'host_add': ATTR.IP(desc='IP of the host the user used to authenticate.'),
        'user_agent': ATTR.STR(
            desc='User-agent of the app the user used to authenticate.'
        ),
        'expiry': ATTR.DATETIME(desc='Python `datetime` ISO format of session expiry.'),
        'token_hash': ATTR.STR(desc='Hashed system-generated session token.'),
        'create_time': ATTR.DATETIME(
            desc='Python `datetime` ISO format of the doc creation.'
        ),
    }
    defaults = {'groups': []}
    extns = {'user': EXTN(module='user', attrs=['*'], force=True)}
    methods = {
        'read': METHOD(
            permissions=[PERM(privilege='read', query_mod={'user': '$__user'})]
        ),
        'create': METHOD(permissions=[PERM(privilege='create')]),
        'update': METHOD(
            permissions=[
                PERM(
                    privilege='update',
                    query_mod={'user': '$__user'},
                    doc_mod={'user': None},
                )
            ],
            query_args={'_id': ATTR.ID()},
        ),
        'delete': METHOD(
            permissions=[PERM(privilege='delete', query_mod={'user': '$__user'})],
            query_args={'_id': ATTR.ID()},
        ),
        'auth': METHOD(permissions=[PERM(privilege='*')], doc_args=[]),
        'reauth': METHOD(
            permissions=[PERM(privilege='*')],
            query_args=[
                {
                    '_id': ATTR.ID(),
                    'token': ATTR.STR(),
                    'groups': ATTR.LIST(list=[ATTR.ID()]),
                },
                {'_id': ATTR.ID(), 'token': ATTR.STR()},
            ],
        ),
        'signout': METHOD(
            permissions=[PERM(privilege='*')],
            query_args={'_id': ATTR.ID()},
        ),
    }

    async def auth(self, skip_events, env, doc):
        for attr in Registry.module('user').unique_attrs:
            if attr in doc.keys():
                key = attr
                break
        user_query = [{key: doc[key], '$limit': 1}]
        if 'groups' in doc.keys():
            user_query.append(
                [{'groups': {'$in': doc['groups']}}, {'privileges': {'*': ['*']}}]
            )
        user_results = await Registry.module('user').read(
            skip_events=[Event.PERM, Event.ON], env=env, query=user_query
        )
        if not user_results['args']['count'] or not pbkdf2_sha512.verify(
            doc['hash'],
            user_results['args']['docs'][0][f'{key}_hash'],
        ):
            raise self.exception(
                status=403,
                msg='Wrong auth credentials.',
                args={'code': 'INVALID_CREDS'},
            )

        user = user_results['args']['docs'][0]

        if Event.ON not in skip_events:
            if user['status'] in ['banned', 'deleted']:
                raise self.exception(
                    status=403,
                    msg=f'User is {user["status"]}.',
                    args={'code': 'INVALID_USER'},
                )

            if user['status'] == 'disabled_password':
                raise self.exception(
                    status=403,
                    msg='User password is disabled.',
                    args={'code': 'INVALID_USER'},
                )

        token = secrets.token_urlsafe(32)
        session = {
            'user': user['_id'],
            'groups': doc['groups'] if 'groups' in doc.keys() else [],
            'host_add': env['REMOTE_ADDR'],
            'user_agent': env['HTTP_USER_AGENT'],
            'expiry': (
                datetime.datetime.utcnow() + datetime.timedelta(days=30)
            ).isoformat(),
            'token_hash': pbkdf2_sha512.using(rounds=100000).hash(token),
        }

        results = await self.create(skip_events=[Event.PERM], env=env, doc=session)
        if results['status'] != 200:
            return results

        session['_id'] = results['args']['docs'][0]['_id']
        session['user'] = user
        del session['token_hash']
        session['token'] = token
        results['args']['docs'][0] = session

        # [DOC] read user privileges and return them
        user_results = await Registry.module('user').read_privileges(
            skip_events=[Event.PERM], env=env, query=[{'_id': user['_id']}]
        )
        if user_results['status'] != 200:
            return user_results
        results['args']['docs'][0]['user'] = user_results['args']['docs'][0]

        # [DOC] Create CONN_AUTH Analytic doc
        if Config.analytics_events['session_conn_auth']:
            analytic_doc = {
                'event': 'CONN_AUTH',
                'subevent': env['client_app'],
                'args': {
                    'user': user_results['args']['docs'][0]['_id'],
                    'session': results['args']['docs'][0]['_id'],
                    'REMOTE_ADDR': env['REMOTE_ADDR'],
                    'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
                },
            }
            analytic_results = await Registry.module('analytic').create(
                skip_events=[Event.PERM], env=env, doc=analytic_doc
            )
            if analytic_results['status'] != 200:
                logger.error(
                    f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
                )
        # [DOC] Create USER_AUTH Analytic doc
        if Config.analytics_events['session_user_auth']:
            analytic_doc = {
                'event': 'USER_AUTH',
                'subevent': user_results['args']['docs'][0]['_id'],
                'args': {
                    'session': results['args']['docs'][0]['_id'],
                    'REMOTE_ADDR': env['REMOTE_ADDR'],
                    'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
                    'client_app': env['client_app'],
                },
            }
            analytic_results = await Registry.module('analytic').create(
                skip_events=[Event.PERM], env=env, doc=analytic_doc
            )
            if analytic_results['status'] != 200:
                logger.error(
                    f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
                )

        return self.status(
            status=200,
            msg='You were successfully authed.',
            args={'session': results['args']['docs'][0]},
        )

    async def reauth(self, env, query):
        if str(query['_id'][0]) == 'f00000000000000000000012':
            raise self.exception(
                status=400,
                msg='Reauth is not required for \'__ANON\' user.',
                args={'code': 'ANON_REAUTH'},
            )

        session_query = [{'_id': query['_id'][0]}]
        if 'groups' in query:
            session_query.append({'groups': {'$in': query['groups'][0]}})
        results = await self.read(
            skip_events=[Event.PERM], env=env, query=session_query
        )
        if not results['args']['count']:
            raise self.exception(
                status=403, msg='Session is invalid.', args={'code': 'INVALID_SESSION'}
            )

        if not pbkdf2_sha512.verify(
            query['token'][0], results['args']['docs'][0]['token_hash']
        ):
            raise self.exception(
                status=403,
                msg='Reauth token hash invalid.',
                args={'code': 'INVALID_REAUTH_HASH'},
            )

        del results['args']['docs'][0]['token_hash']
        results['args']['docs'][0]['token'] = query['token'][0]

        if results['args']['docs'][0]['expiry'] < datetime.datetime.utcnow().isoformat():
            results = await self.delete(
                skip_events=[Event.PERM, Event.SOFT],
                env=env,
                query=[{'_id': env['session']['_id']}],
            )
            raise self.exception(
                status=403, msg='Session had expired.', args={'code': 'SESSION_EXPIRED'}
            )

        # [DOC] update user's last_login timestamp
        await Registry.module('user').update(
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': results['args']['docs'][0]['user']}],
            doc={'login_time': datetime.datetime.utcnow().isoformat()},
        )
        await self.update(
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': results['args']['docs'][0]['_id']}],
            doc={
                'expiry': (
                    datetime.datetime.utcnow() + datetime.timedelta(days=30)
                ).isoformat()
            },
        )
        # [DOC] read user privileges and return them
        user_results = await Registry.module('user').read_privileges(
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': results['args']['docs'][0]['user']['_id']}],
        )
        results['args']['docs'][0]['user'] = user_results['args']['docs'][0]

        # [DOC] Create CONN_AUTH Analytic doc
        if Config.analytics_events['session_conn_reauth']:
            analytic_doc = {
                'event': 'CONN_REAUTH',
                'subevent': env['client_app'],
                'args': {
                    'user': user_results['args']['docs'][0]['_id'],
                    'session': results['args']['docs'][0]['_id'],
                    'REMOTE_ADDR': env['REMOTE_ADDR'],
                    'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
                },
            }
            analytic_results = await Registry.module('analytic').create(
                skip_events=[Event.PERM], env=env, doc=analytic_doc
            )
            if analytic_results['status'] != 200:
                logger.error(
                    f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
                )
        # [DOC] Create USER_AUTH Analytic doc
        if Config.analytics_events['session_user_reauth']:
            analytic_doc = {
                'event': 'USER_REAUTH',
                'subevent': user_results['args']['docs'][0]['_id'],
                'args': {
                    'session': results['args']['docs'][0]['_id'],
                    'REMOTE_ADDR': env['REMOTE_ADDR'],
                    'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
                    'client_app': env['client_app'],
                },
            }
            analytic_results = await Registry.module('analytic').create(
                skip_events=[Event.PERM], env=env, doc=analytic_doc
            )
            if analytic_results['status'] != 200:
                logger.error(
                    f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
                )

        return self.status(
            status=200,
            msg='You were successfully reauthed.',
            args={'session': results['args']['docs'][0]},
        )

    async def signout(self, env, query):
        if str(query['_id'][0]) == 'f00000000000000000000012':
            raise self.exception(
                status=400,
                msg='Singout is not allowed for \'__ANON\' user.',
                args={'code': 'ANON_SIGNOUT'},
            )

        results = await self.read(
            skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
        )

        if not results['args']['count']:
            raise self.exception(
                status=403, msg='Session is invalid.', args={'code': 'INVALID_SESSION'}
            )

        results = await self.delete(
            skip_events=[Event.PERM], env=env, query=[{'_id': env['session']['_id']}]
        )

        # [DOC] Create CONN_AUTH Analytic doc
        if Config.analytics_events['session_conn_deauth']:
            analytic_doc = {
                'event': 'CONN_DEAUTH',
                'subevent': env['client_app'],
                'args': {
                    'user': env['session']['user']['_id'],
                    'session': env['session']['_id'],
                    'REMOTE_ADDR': env['REMOTE_ADDR'],
                    'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
                },
            }
            analytic_results = await Registry.module('analytic').create(
                skip_events=[Event.PERM], env=env, doc=analytic_doc
            )
            if analytic_results['status'] != 200:
                logger.error(
                    f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
                )
        # [DOC] Create USER_AUTH Analytic doc
        if Config.analytics_events['session_user_deauth']:
            analytic_doc = {
                'event': 'USER_DEAUTH',
                'subevent': env['session']['user']['_id'],
                'args': {
                    'session': env['session']['_id'],
                    'REMOTE_ADDR': env['REMOTE_ADDR'],
                    'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
                    'client_app': env['client_app'],
                },
            }
            analytic_results = await Registry.module('analytic').create(
                skip_events=[Event.PERM], env=env, doc=analytic_doc
            )
            if analytic_results['status'] != 200:
                logger.error(
                    f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
                )

        return self.status(
            status=200,
            msg='You are successfully signed-out.',
            args={'session': {'_id': 'f00000000000000000000012'}},
        )



import asyncio
import copy
import datetime
import inspect
import json
import logging
from typing import TYPE_CHECKING, Optional, cast

import jwt
import redis
from bson import ObjectId

from nawah.classes import JSONEncoder
from nawah.config import Config
from nawah.enums import Event

if TYPE_CHECKING:
    from nawah.base_module import BaseModule
    from nawah.classes import CACHE, NAWAH_ENV, NAWAH_EVENTS, NAWAH_QUERY

logger = logging.getLogger('nawah')


def _generate_cache_key(
    cache: 'CACHE',
    method: str,
    module: 'BaseModule',
    skip_events: 'NAWAH_EVENTS',
    env: 'NAWAH_ENV',
    query: 'NAWAH_QUERY',
) -> Optional[str]:
    if not Config._sys_cache or Event.CACHE in skip_events or not cache:
        return None

    condition_params = {
        'skip_events': skip_events,
        'env': env,
        'query': query,
    }

    if not cache.condition(
        **{
            param: condition_params[param]
            for param in inspect.signature(cache.condition).parameters
        }
    ):
        return None

    cache_key = {
        'query': JSONEncoder().encode(query._query),
        'special': JSONEncoder().encode(query._special),
        'extn': Event.EXTN in skip_events,
        'user': env['session']['user']['_id'] if cache.user_scoped else None,
    }

    cache_key_jwt = jwt.encode(cache_key, '_').split('.')[1]

    return cache_key_jwt


def _call_cache_decoder(cache_dict):
    if not cache_dict:
        return

    for key in cache_dict:
        if isinstance(cache_dict[key], dict):
            if '$oid' in cache_dict[key]:
                cache_dict[key] = ObjectId(cache_dict[key]['$oid'])
            else:
                _call_cache_decoder(cache_dict=cache_dict[key])
        elif isinstance(cache_dict[key], list):
            for i in range(len(cache_dict[key])):
                if isinstance(cache_dict[key][i], dict):
                    if '$oid' in cache_dict[key][i]:
                        cache_dict[key][i] = ObjectId(cache_dict[key][i]['$oid'])
                    else:
                        _call_cache_decoder(cache_dict=cache_dict[key][i])


def _call_cache(cache: 'CACHE', method: str, module: 'BaseModule', cache_key: str):
    if not Config._sys_cache:
        return

    try:
        logger.debug(
            'Attempting to get cache with \'key\': \'%s\'.',
            f'.{module.module_name}.{method}.{cache_key}',
        )
        cache_dict = Config._sys_cache.get(
            cache.channel,
            f'.{module.module_name}.{method}.{cache_key}.results',
        )

        _call_cache_decoder(cache_dict=cache_dict)

        return cache_dict

    except redis.exceptions.ResponseError:
        return

    except redis.exceptions.ConnectionError:
        logger.error(
            'Connection with Redis server \'%s\' failed. Skipping Cache Workflow.',
            Config.cache_server,
        )
        return


async def _set_cache(channel, module, method, cache_key, results):
    if not Config._sys_cache:
        return

    results = copy.deepcopy(results)

    try:
        logger.debug(
            'Attempting to set cache with \'key\': \'%s\'.',
            f'.{module}.{method}.{cache_key}',
        )

        if not Config._sys_cache.get(channel):
            Config._sys_cache.set(channel, '.', {})
        try:
            Config._sys_cache.get(channel, f'.{module}')
        except redis.exceptions.ResponseError:
            Config._sys_cache.set(channel, f'.{module}', {})
        try:
            Config._sys_cache.get(channel, f'.{module}.{method}')
        except redis.exceptions.ResponseError:
            Config._sys_cache.set(
                channel,
                f'.{module}.{method}',
                {},
            )

        results = {
            'docs': [doc['_id'] for doc in results['args']['docs']]
            if 'args' in results and 'docs' in results['args']
            else [],
            'results': results,
        }

        Config._sys_cache.set(
            channel,
            f'{module}.{method}.{cache_key}',
            json.loads(JSONEncoder(object_id_dict=True).encode(results)),
        )
    except redis.exceptions.ConnectionError:
        logger.error(
            'Connection with Redis server \'%s\' failed. Skipping Cache Workflow.',
            Config.cache_server,
        )


def _call_method(
    cache: 'CACHE',
    method: str,
    module: 'BaseModule',
    skip_events: 'NAWAH_EVENTS',
    env: 'NAWAH_ENV',
    query: 'NAWAH_QUERY',
):
    cache_key = _generate_cache_key(
        cache=cache,
        method=method,
        module=module,
        skip_events=skip_events,
        env=env,
        query=query,
    )
    call_cache = None
    if cache_key:
        cache_key = cast(str, cache_key)
        call_cache = _call_cache(
            cache=cache, method=method, module=module, cache_key=cache_key
        )

    async def _method(skip_events, env, query, doc):
        if call_cache:
            return call_cache

        module_method = getattr(module, f'_method_{method}')

        method_params = {
            'skip_events': skip_events,
            'env': env,
            'query': query,
            'doc': doc,
        }

        results = await module_method(
            **{
                param: method_params[param]
                for param in inspect.signature(module_method).parameters
            }
        )

        if cache_key:
            results['args']['cache_key'] = cache_key
            if 'cache_time' not in results['args']:
                logger.debug(
                    'Results generated with \'cache_key\'. Calling \'_set_cache\'.'
                )
                results['args']['cache_time'] = datetime.datetime.utcnow().isoformat()
                asyncio.create_task(
                    _set_cache(
                        cache.channel, module.module_name, method, cache_key, results
                    )
                )

        return results

    return _method
